<?php
namespace Civi\Updaterecur;

use Civi\Api4\Contribution;
use Civi\Api4\ContributionRecur;
use CRM_Updaterecur_ExtensionUtil as E;

class Utils {
  const EVENT_GIVING_DETAILS = 'civi.updaterecur.givingdetails';

  public static function registerTokens(\Civi\Token\Event\TokenRegisterEvent $e) {
    $e->entity('contact')
      ->register('currentRegularGiving', ts('List all current regular giving (Recurring Contributions)'));
  }

  public static function evaluateTokens(\Civi\Token\Event\TokenValueEvent $e) {

    foreach ($e->getRows() as $row) {
      /** @var \Civi\Token\TokenRow $row */
      $row->format('text/html');
      $includeTestData = TRUE;
      $details = Utils::getGivingDetails($row->context['contactId'], $includeTestData);
      if (count($details['descriptions']) === 1) {
        $description = '<p>' . reset($details['descriptions']) . '</p>';
      }
      elseif (count($details['descriptions']) > 1) {
        $description = E::ts("<p>It looks like you have several separate regular donations set up:</p><ul><li>")
          . implode("</li><li>", $details['descriptions'])
          . '</li></ul>';
      }
      else {
        $description = '<p>' . E::ts("You don’t have any regular donations set up.") . '</p>';
      }
      $row->tokens('contact', 'currentRegularGiving', $description ?? '');
    }

  }

  /**
   * @return array
   *   This contains:
   *   - 'crs' array of current contribution recurs
   *   - 'description' HTML description of giving.
   *   - 'contactID'
   *
   *   Note that later, the inlay may add, or these could be added by the hook/event.
   *   - 'suggested' Amount to suggest, if a change looks possible.
   *   - 'updateDirectly' bool, whether we can attempt a change directly (e.g. single GC DD)
   *   - 'currentAmount' set if a single cr is found
   *
   */
  public static function getGivingDetails(int $contactID, bool $includeTestData = FALSE): array {
    $crs = ContributionRecur::get(FALSE)
      ->addWhere('contact_id', '=', $contactID)
      ->addWhere('is_test', 'IN', $includeTestData ? [0, 1] : [0])
      ->addWhere('contribution_status_id:name', '=', 'In Progress')
      ->addSelect('amount', 'currency',
        'frequency_unit:label', /*month*/
        'frequency_interval',
        'payment_processor_id',
        'payment_processor_id.payment_processor_type_id:name',
        'payment_instrument_id:label',
        'financial_type_id:name',
        'financial_type_id:label',
        'is_test',
      )
      ->execute()->getArrayCopy();

    foreach ($crs as &$cr) {
      $cr['lastCn'] = Contribution::get(FALSE)
        ->addWhere('contribution_recur_id', '=', $cr['id'])
        ->addWhere('is_test', 'IN', $includeTestData ? [0, 1] : [0])
        ->addWhere('is_template', '=', FALSE)
        ->addSelect('total_amount', 'receive_date', 'is_test')
        ->addOrderBy('receive_date', 'DESC')
        ->setLimit(1)->execute()->first();
    }
    unset($cr);

    $descriptions = array_map('\\Civi\Updaterecur\\Utils::summariseCr', $crs);
    $return = ['crs' => $crs] + compact('descriptions', 'contactID');

    // Allow local code to overrule our findings.
    $event = \Civi\Core\Event\GenericHookEvent::create(['data' => &$return]);
    \Civi::dispatcher()->dispatch(self::EVENT_GIVING_DETAILS, $event);

    return $return;
  }

  /**
   * Returns HTML summary of a recur, and last payment date if present.
   */
  public static function summariseCr(array $cr): string {
    $description = 'You are giving <strong>' . htmlspecialchars(static::summariseCrAmountAndInterval($cr)) . '</strong>';
    if (!empty($cr['lastCn'])) {
      $description .= ' Last payment on ' . date('j M Y', strtotime($cr['lastCn']['receive_date'])) . ', thank you.';
    }
    return $description;
  }

  /**
   * Returns plain text summary of a recur.
   * e.g. "£5.00 every month"
   */
  public static function summariseCrAmountAndInterval(array $cr): string {
    $symbol = ['GBP' => '£', 'USD' => '$', 'CAD' => '$', 'EUR' => '€'][$cr['currency']] ?? $cr['currency'];
    $n = ($cr['frequency_interval'] == 1) ? '' : $cr['frequency_interval'];
    return "$symbol{$cr['amount']} every $n {$cr['frequency_unit:label']} by "
        . $cr['payment_instrument_id:label'] . '.';
  }

}
