<?php

use Civi\Test\HeadlessInterface;
use Civi\Test\HookInterface;
use Civi\Test\TransactionalInterface;
use Civi\Payment\Exception\PaymentProcessorException;

/**
 * Job.Performsubscriptionchanges API Test Case
 * This is a generic test class implemented with PHPUnit.
 * @group headless
 */
class api_v3_Job_PerformsubscriptionchangesTest extends \PHPUnit\Framework\TestCase implements HeadlessInterface, HookInterface, TransactionalInterface {
  use \Civi\Test\Api3TestTrait;
  use \Civi\UpdateRecurTraits;

  protected $tplContribIdByRecurId = [];
  protected $extraPriceVieldValueID = 0;

  /**
   * Set up for headless tests.
   *
   * Civi\Test has many helpers, like install(), uninstall(), sql(), and sqlFile().
   *
   * See: https://docs.civicrm.org/dev/en/latest/testing/phpunit/#civitest
   */
  public function setUpHeadless() {
    return \Civi\Test::headless()
      ->install(['org.civicrm.search_kit', 'org.civicrm.afform'])
      ->installMe(__DIR__)
      ->apply();
  }

  /**
   * The setup() method is executed before the test is executed (optional).
   */
  public function setUp() :void {
    parent::setUp();
  }

  /**
   * The tearDown() method is executed after the test was executed (optional)
   * This can be used for cleanup.
   */
  public function tearDown() :void {

    if ($this->pps) {
      \Civi\Api4\ContributionRecur::delete(FALSE)
        ->addWhere('payment_processor_id', 'IN', array_values($this->pps))
        ->execute();
      ;
      \Civi\Api4\PaymentProcessor::delete(FALSE)
        ->addWhere('id', 'IN', array_values($this->pps))
        ->execute();
      ;
    }

    \Civi\Api4\PaymentProcessorType::delete(FALSE)
      ->addWhere('name', 'IN', [
        'DummyProcessorNoSupport',
        'DummyProcessorOldAmountUnsupported',
        'DummyProcessorOldAmountSupported',
        'DummyProcessorOldAmountSupportedButWillFail',
        'DummyProcessorNewAmountDirect',
        'DummyProcessorNewAmountForbidden',
        'DummyProcessorNewAmountSupported',
        'DummyProcessorNewAmountSupportedButWillFail',
      ])
      ->execute();
    ;
    parent::tearDown();
  }

  /**
   * Test that CRs without a payment processor get updated OK.
   *
   * No template contribution to update.
   */
  public function testWorksWithoutPaymentProcessor() {

    list($performResult, $activities, $recurs) = $this->commonSetup(NULL, 2);

    // Without a processor we expect them both to succeed.
    $this->assertEquals(['success' => 2, 'fail' => 0, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Completed', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@
      There.is.no.template.contribution.assigned.to.Recurring.Contribution.ID.{$this->recurs[$this->contacts['WilmaUpgrader']]}.so.nothing.to.update.*?
      Successfully.updated.*?Recurring.contribution.ID.{$this->recurs[$this->contacts['WilmaUpgrader']]}.was.updated.without.using.the.payment.processor
      @sx", $wilmaUpgrader['details']);

    $barneyDowngrader = $activities[$this->contacts['BarneyDowngrader']];
    $this->assertEquals('Completed', $barneyDowngrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Successfully updated.*Recurring contribution ID {$this->recurs[$this->contacts['BarneyDowngrader']]} was updated without using the payment processor.@", $barneyDowngrader['details']);

    $this->assertEquals(2, $recurs[$this->contacts['WilmaUpgrader']]['amount']);
    $this->assertEquals(2, $recurs[$this->contacts['BarneyDowngrader']]['amount']);
  }

  /**
   * Test that CRs without a payment processor get updated OK.
   *
   * No template contribution to update.
   */
  public function testTemplateContributionsChanges() {

    $n = 1;
    $paymentProcessorName = NULL;
    $withTplContrib = 'simple';
    list($performResult, $activities, $recurs) = $this->commonSetup($paymentProcessorName, $n, $withTplContrib);
    $contactID = $this->contacts['WilmaUpgrader'];
    $recurID = $this->recurs[$contactID];
    $tplContribID = $this->tplContribIdByRecurId[$recurID];

    // Should succeed (this is a repeat of other tests)
    $this->assertEquals(['success' => 1, 'fail' => 0, 'pending' => 0], $performResult['values']);
    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Completed', $wilmaUpgrader['status_id:name']);
    $this->assertEquals(2, $recurs[$this->contacts['WilmaUpgrader']]['amount']);

    // Now check tpl contribution was updated.
    $this->assertMatchesRegularExpression("@
      Updated.line.item.on.template.contribution.*?
      Updated.template.contribution:.Template.Contribution.found,.ID.$tplContribID,.set.amount.to.2.*?
      Successfully.updated.*?Recurring.contribution.ID.{$recurID}.was.updated.without.using.the.payment.processor
      @sx", $wilmaUpgrader['details']);

  }

  /**
   */
  public function testTemplateContributionWithTooManyLineItemsFails() {

    $n = 1;
    $paymentProcessorName = NULL;
    $withTplContrib = 'complex';
    list($performResult, $activities, $recurs) = $this->commonSetup($paymentProcessorName, $n, $withTplContrib);
    $contactID = $this->contacts['WilmaUpgrader'];
    $recurID = $this->recurs[$contactID];
    $tplContribID = $this->tplContribIdByRecurId[$recurID];

    // Should fail (this is a repeat of other tests)
    $this->assertEquals(['success' => 0, 'fail' => 1, 'pending' => 0], $performResult['values']);
    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Cancelled', $wilmaUpgrader['status_id:name']);
    $this->assertEquals(2, $recurs[$this->contacts['WilmaUpgrader']]['amount']);

    $this->assertMatchesRegularExpression("@
      Updated.Recurring.Contribution.ID.$recurID,.set.amount.to.2.*?
      Failed.to.change.*?
      The.template.contribution.\(id.\d+\).for.recurring.contribution.ID.$recurID.includes.2.line.items
      @sx", $wilmaUpgrader['details']);

  }

  /**
   * Test with dummy processor type that has no explicit support at all (it's just vanilla CRM_Core_Payment)
   *
   * We expect the records to be updated directly.
   *
   * This uses two contacts, just to check the loop works.
   */
  public function testWorksWithDummyProcessorNoSupport() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorNoSupport', 2);

    // Without a processor we expect them both to succeed.
    $this->assertEquals(['success' => 2, 'fail' => 0, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Completed', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Successfully updated.*Recurring contribution ID {$this->recurs[$this->contacts['WilmaUpgrader']]} was updated without using the payment processor.@", $wilmaUpgrader['details']);

    $this->assertEquals(2, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount']);
  }

  /**
   * Test with dummy processor type that supportsChangeSubscriptionAmount() by
   * implementing changeSubscriptionAmount()
   *
   * We expect the records to be updated via the processor.
   */
  public function testWorksWithDummyProcessorOldAmountSupportedWillSucceed() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorOldAmountSupported');

    $this->assertEquals(['success' => 1, 'fail' => 0, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Completed', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Successfully updated.*?Recurring contribution ID {$this->recurs[$this->contacts['WilmaUpgrader']]} was updated by the payment processor \(changeSubscriptionAmount\): External processor accepted update@", $wilmaUpgrader['details']);

    $this->assertEquals(2, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount']);
  }

  /**
   * Test with dummy processor type that supportsChangeSubscriptionAmount() by
   * implementing changeSubscriptionAmount() but throws a PaymentProcessorException
   *
   * We expect the records to be updated via the processor.
   */
  public function testWorksWithDummyProcessorOldAmountSupportedButWillFail() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorOldAmountSupportedButWillFail');

    // Without a processor we expect them both to succeed.
    $this->assertEquals(['success' => 0, 'fail' => 1, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Cancelled', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Change rejected by payment processor.*External processor rejected update@s", $wilmaUpgrader['details']);

    $this->assertEquals(1, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount'],
      "Expect amount to be as-was");
  }

  /**
   * Test with dummy processor type that supports future way but whose getSupportedSubscriptionChanges
   * says that Civi's free to do what it wants with amount changes without needing the processor.
   */
  public function testWorksWithDummyProcessorNewAmountDirect() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorNewAmountDirect');

    $this->assertEquals(['success' => 1, 'fail' => 0, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Completed', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Successfully updated.*Recurring contribution ID {$this->recurs[$this->contacts['WilmaUpgrader']]} was updated without using the payment processor.@", $wilmaUpgrader['details']);

    $this->assertEquals(2, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount']);
  }

  /**
   * Test with dummy processor type that supports future way but whose getSupportedSubscriptionChanges
   * says that changes to amount are not permitted.
   */
  public function testWorksWithDummyProcessorNewAmountForbidden() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorNewAmountForbidden');

    $this->assertEquals(['success' => 0, 'fail' => 1, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Cancelled', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@The payment processor \('DummyProcessorNewAmountForbidden' ID \d+\) forbids changing the amount. Cannot process recurring Contribution \(ID \d+\)@", htmlspecialchars_decode($wilmaUpgrader['details']));

    $this->assertEquals(1, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount']);
  }

  /**
   * Test with dummy processor type that supports future way via changeSubscriptions
   */
  public function testWorksWithDummyProcessorNewAmountSupportedWillSucceed() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorNewAmountSupported');

    $this->assertEquals(['success' => 1, 'fail' => 0, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Completed', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Successfully updated.*?Recurring contribution ID \d+ was updated by the payment processor \(changeSubscription\)@", $wilmaUpgrader['details']);

    $this->assertEquals(2, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount']);
  }

  /**
   * Test with dummy processor type that supports future way but whose changeSubscription method
   * rejects the request.
   */
  public function testWorksWithDummyProcessorNewAmountSupportedButWillFail() {
    list($performResult, $activities, $recursByContactID) = $this->commonSetup('DummyProcessorNewAmountSupportedButWillFail');

    $this->assertEquals(['success' => 0, 'fail' => 1, 'pending' => 0], $performResult['values']);

    $wilmaUpgrader = $activities[$this->contacts['WilmaUpgrader']];
    $this->assertEquals('Cancelled', $wilmaUpgrader['status_id:name']);
    $this->assertMatchesRegularExpression("@Change rejected by payment processor.*?External processor rejected changing amount@", $wilmaUpgrader['details']);

    $this->assertEquals(1, $recursByContactID[$this->contacts['WilmaUpgrader']]['amount']);
  }

  protected function createPaymentProcessorFixtures() {
    $processors = [
      'DummyProcessorNoSupport',
      'DummyProcessorOldAmountUnsupported',
      'DummyProcessorOldAmountSupported',
      'DummyProcessorOldAmountSupportedButWillFail',
      'DummyProcessorNewAmountDirect',
      'DummyProcessorNewAmountForbidden',
      'DummyProcessorNewAmountSupported',
      'DummyProcessorNewAmountSupportedButWillFail',
    ];
    $records = [];
    foreach ($processors as $processorClass) {
      $records[] = [
        'name' => $processorClass,
        'title' => $processorClass,
        'class_name' => $processorClass,
      ];
    }
    $ppTypesToIDs = \Civi\Api4\PaymentProcessorType::save(FALSE)
      ->setDefaults([
        'is_active' => TRUE,
        'billing_mode' => 1,
      ])
      ->setRecords($records)
      ->execute()
      ->indexBy('name')->column('id');

    // Create (test) instances.
    $records = [];
    foreach ($ppTypesToIDs as $class => $ppTypeID) {
      $records[] = [
        'name' => "$class",
        'payment_processor_type_id' => $ppTypeID,
      ];
    }
    $this->pps = \Civi\Api4\PaymentProcessor::save(FALSE)
      ->setDefaults([
        'is_active' => TRUE,
        'is_test' => TRUE,
        'financial_account_id.name' => 'Deposit Bank Account',
      ])
      ->setRecords($records)
      ->execute()
      ->indexBy('name')->column('id');

  }

  /**
   * @var mixed $withTplContrib
   *    - FALSEy: no template contrib.
   *    - string 'simple': template + single line item.
   *    - string 'complex': template + multiple line items (cannot be updated)
   */
  protected function commonSetup(?string $ppTypeName = NULL, $n = 1, $withTplContrib = FALSE) {
    $this->createPaymentProcessorFixtures();

    // Create a 2nd price_field_value_id on the default price set.
    $orig = \Civi\Api4\PriceFieldValue::get(FALSE)
      ->addWhere('id', '=', 1)
      ->addWhere('price_field_id', '=', 1)
      ->execute()->first();
    unset($orig['id']);
    $orig['label'] = 'phpunit test purposes';
    $this->extraPriceVieldValueID = (int) \Civi\Api4\PriceFieldValue::create(FALSE)
      ->setValues($orig)->execute()->first()['id'] ?? 0;

    // Create the fixture inc. $n change activities (this is tested separately in ChangeSubscriptionTest)
    $this->createFixture($ppTypeName, $n);

    if ($withTplContrib) {
      $this->addTemplateContribution($withTplContrib);
    }

    $activities = \Civi\Api4\ContributionRecur::changeSubscriptions(FALSE)
      ->setWhere([['id', 'IN', array_values($this->recurs)]])
      ->setNewAmount('2')
      ->setReason('test change to 2 GBP')
      ->execute();

    // The activities will need their dates resetting so that they are processed.
    \Civi\Api4\Activity::update(FALSE)
      ->addWhere('id', 'IN', array_values($activities->column('id')))
      ->addValue('activity_date_time', date('Y-m-d H:i:s', time() - 60))
      ->execute();

    // Process the changes.
    $performResult = civicrm_api3('Job', 'Performsubscriptionchanges', ['time_limit' => 0]);

    $recursByContactID = \Civi\Api4\ContributionRecur::get(FALSE)
      ->addWhere('id', 'IN', array_values($this->recurs))
      ->addSelect('id', 'amount', 'contact_id')
      ->execute()->indexBy('contact_id')->getArrayCopy();

    // Find the activities.
    $activities = \Civi\Api4\Activity::get(FALSE)
      ->addJoin('ActivityContact AS target', 'INNER', ['target.record_type_id', '=', 3])
      ->addSelect(
        'activity_type_id:name',
        'activity_date_time',
        'status_id:name',
        'subject',
        'details',
        'updaterecur_activity_fields.updaterecur_recur_id',
        'updaterecur_activity_fields.updaterecur_old_amount',
        'updaterecur_activity_fields.updaterecur_new_amount',
        'target.contact_id'
      )
      ->addWhere('id', 'IN', $activities->column('id'))
      ->execute()
      ->indexBy('target.contact_id')->getArrayCopy();

    return [$performResult, $activities, $recursByContactID];
  }

  /**
   * This copies the pattern from CRM_Contribute_BAO_ContributionRecur::ensureTemplateContributionExists()
   */
  protected function addTemplateContribution(string $type) {
    foreach ($this->contacts as $name => $contactID) {
      $recurID = $this->recurs[$contactID];

      // Create a template contribution against the recur.
      $templateContributionParams = [
        'contact_id' => $contactID,
        'contribution_recur_id' => $recurID,
        'is_test' => 1,
        'is_template' => 1,
        'skipRecentView' => TRUE,
        'status_id' => 'Template',
        'financial_type_id' => 'Donation',
        'currency' => 'GBP',
        'source_contact_id' => $contactID,
      // WilmaUpgrader is currently paying £1
        'total_amount' => 1,
      ];
      $templateContribution = civicrm_api3('Contribution', 'create', $templateContributionParams);
      // Add new soft credit against current $contribution.
      // artfulrobot: I do NOT know what this does or why it's useful.
      CRM_Contribute_BAO_ContributionRecur::addrecurSoftCredit($templateContributionParams['contribution_recur_id'], $templateContribution['id']);
      $this->tplContribIdByRecurId[$recurID] = (int) $templateContribution['id'];

      $tplContribID = \Civi\Api4\Contribution::get(FALSE)
        ->addWhere('contribution_recur_id', '=', $recurID)
        ->addWhere('is_template', '=', 1)
        ->addWhere('is_test', '=', 1)
        ->addSelect('id')
        ->execute()->first()['id'] ?? 0;
      $this->assertGreaterThan(0, $tplContribID);

      if ($type === 'complex') {
        // Load the line item,
        $lineItem = \Civi\Api4\LineItem::get(FALSE)
          ->addWhere('contribution_id', '=', $tplContribID)
          ->execute()->first();
        // Add a new line item, with a zero amount.
        unset($lineItem['id']);
        $lineItem['unit_price'] = 0;
        $lineItem['line_total'] = 0;
        $lineItem['price_field_value_id'] = $this->extraPriceVieldValueID;

        \Civi\Api4\LineItem::save(FALSE)
          ->setRecords([$lineItem])
          ->execute();
      }
    }
  }

}
class DummyProcessorBase extends CRM_Core_Payment {

  public function checkConfig() {}

}
class CRM_Core_DummyProcessorNoSupport extends DummyProcessorBase {
}
class CRM_Core_DummyProcessorOldAmountSupported extends DummyProcessorBase {

  public function changeSubscriptionAmount(string &$message, array $params) {
    $message = "External processor accepted update";
    return TRUE;
  }

}
class CRM_Core_DummyProcessorOldAmountSupportedButWillFail extends DummyProcessorBase {

  public function changeSubscriptionAmount(string &$message, array $params) {
    $message = "External processor rejected update";
    throw new PaymentProcessorException("$message");
  }

}
class CRM_Core_DummyProcessorNewAmountDirect extends DummyProcessorBase {

  /**
   * New proposed method
   */
  public function getSupportedSubscriptionChanges() {
    // by not returning anything for 'amount' Civi should revert to doing it's normal direct update.
    return [];
  }

}
class CRM_Core_DummyProcessorNewAmountForbidden extends DummyProcessorBase {

  /**
   * New proposed method
   */
  public function getSupportedSubscriptionChanges() {
    // by returning FALSE for amount, we're saying that we don't allow it.
    return ['amount' => FALSE];
  }

}
class CRM_Core_DummyProcessorNewAmountSupported extends DummyProcessorBase {

  /**
   * New proposed method
   */
  public function getSupportedSubscriptionChanges() {
    // by returning FALSE for amount, we're saying that we don't allow it.
    return ['amount' => TRUE];
  }

  public function changeSubscription(int $recurID, array $changes) {
    // We take responsibility for updating the ContributionRecur.
    \Civi\Api4\ContributionRecur::update(FALSE)
      ->addValue('amount', $changes['amount'])
      ->addWhere('id', '=', $recurID)
      ->execute();
  }

}
class CRM_Core_DummyProcessorNewAmountSupportedButWillFail extends DummyProcessorBase {

  /**
   * New proposed method
   */
  public function getSupportedSubscriptionChanges() {
    // by returning FALSE for amount, we're saying that we don't allow it.
    return ['amount' => TRUE];
  }

  public function changeSubscription(int $recurID, array $changes) {
    throw new PaymentProcessorException("External processor rejected changing amount");
  }

}
